import click
from rich.console import Console

from ..commands import render, render_config, single_render
from ..config import GITHUB_TOKEN
from ..errors import InvalidGithubToken
from ..metrics import repository_metrics
from ..version import __version__

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, "-v", "--version", message="version %(version)s")
def cli() -> None:
    """Reviews - A terminal UI Dashboard for monitoring code review requests.\n

    For feature requests or bug reports: https://github.com/apoclyps/reviews/issues
    """


@cli.command(help="Show the activity metrics per repository")
def metrics() -> None:
    """
    Command:\n
        reviews metrics
    """
    repository_metrics()


@cli.command(help="Show the current configuration used by Reviews")
@click.option("-show", "--show/--hide", default=False, is_flag=True)
def config(show: bool) -> None:
    """
    Command:\n
        reviews config

    Usage:\n
        reviews config --show \n
        reviews config --hide \n
    """
    render_config(show=show)


@cli.command(help="Visualize code review requests as a Dashboard")
@click.option("-p", "--provider", type=str, default="github", show_default="github")
@click.option("-r", "--reload/--no-reload", default=True, is_flag=True)
def dashboard(reload: bool, provider: str) -> None:
    """
    Command:\n
        reviews dashboard

    Usage:\n
        reviews dashboard --reload \n
        reviews dashboard --no-reload \n
        reviews dashboard --provider=github \n
        reviews dashboard --provider=gitlab \n
    """
    console = Console()
    style = "bold"

    github_token_msg = (
        "> a valid GITHUB_TOKEN should be provided as an environmental "  # nosec
        "variable or a value within a settings.ini or .env file. Please see "
        "the project [link=https://github.com/apoclyps/reviews]"
        "README[/link] for more methods on providing configuration."
        "\n"
    )

    if not GITHUB_TOKEN:
        missing_github_token_msg = (
            "\n[red]GITHUB_TOKEN[/] is required and has not been provided as "
            "configuration needed to use Reviews. \n\n"
            f"{github_token_msg}"
        )
        console.print(missing_github_token_msg, style=style)
        return

    provider = provider.lower()
    if provider not in ["github", "gitlab"]:
        invalid_providern_msg = (
            "\n[red]--provider=[/] uses an unsupported value. "
            "Please provider either `github` or `gitlab` as the provider value. \n\n"
            f"{github_token_msg}"
        )
        console.print(invalid_providern_msg, style=style)
        return

    try:
        click.echo(f"loading dashboard for {provider}...")

        if reload:
            render(provider=provider)
        else:
            single_render(provider=provider)
    except InvalidGithubToken:
        invalid_github_token_msg = (
            "\n[red]GITHUB_TOKEN[/] is required configuration and an invalid "
            "token has been supplied to Reviews. \n\n"
            f"{github_token_msg}"
        )
        console.print(
            invalid_github_token_msg,
            style=style,
        )

    return


def main() -> None:
    """Entry point to CLI"""
    cli()



import os
from unittest import mock

from reviews import config


@mock.patch.dict(os.environ, {"REVIEWS_GITHUB_REPOSITORY_CONFIGURATION": "apoclyps/reviews"})
def test_reviews_github_repository_configuration():
    configuration = config.get_configuration(config=config.REVIEWS_GITHUB_REPOSITORY_CONFIGURATION)

    assert configuration == [
        ("apoclyps", "reviews"),
    ]


@mock.patch.dict(os.environ, {"REVIEWS_GITLAB_REPOSITORY_CONFIGURATION": "27629846:apoclyps/reviews"})
def test_reviews_gitlab_repository_configuration():
    configuration = config.get_configuration(config=config.REVIEWS_GITLAB_REPOSITORY_CONFIGURATION)

    assert configuration == [
        ("27629846", "apoclyps/reviews"),
    ]



from typing import List, Tuple

from decouple import Csv


def get_configuration(config: Csv) -> List[Tuple[str, str]]:
    """converts a comma separated list of organizations/repositories into a list
    of tuples.
    """

    def _to_tuple(values: List[str]) -> Tuple[str, str]:
        return (values[0], values[1])

    def split(configuration: str) -> List[str]:
        if ":" in configuration:
            return configuration.split(sep=":", maxsplit=1)

        return configuration.split(sep="/", maxsplit=1)

    return [_to_tuple(values=split(configuration=configuration)) for configuration in config]



from typing import List

from github import Github
from github.GithubException import BadCredentialsException, UnknownObjectException
from github.PullRequest import PullRequest
from github.Repository import Repository
from gitlab import Gitlab
from gitlab.v4.objects import ProjectMergeRequest as GitlabMergeRequest
from gitlab.v4.objects.projects import Project as GitlabRepository

from .. import config
from ..errors import InvalidGithubToken, InvalidGitlabToken, RepositoryDoesNotExist


class GithubAPI:
    """Create and execute requests using the Github API"""

    def __init__(self) -> None:
        self._client = Github(
            config.GITHUB_TOKEN,
            per_page=config.GITHUB_DEFAULT_PAGE_SIZE,
            base_url=config.GITHUB_URL,
        )

    def get_repository(self, org: str, repo: str) -> Repository:
        """Returns a repository for a given organization."""
        try:
            return self._client.get_repo(f"{org}/{repo}")
        except UnknownObjectException:
            raise RepositoryDoesNotExist(f"{org}/{repo} does not exist")
        except BadCredentialsException:
            raise InvalidGithubToken("GITHUB_TOKEN is not valid.")

    @staticmethod
    def _get_pull_requests(
        repository: Repository,
        state: str = "open",
        sort: str = "created",
    ) -> List[PullRequest]:
        return list(repository.get_pulls(state=state, sort=sort))

    def get_pull_requests(self, org: str, repo: str) -> List[PullRequest]:
        """Returns a list of pull requests for a given organization and repository."""
        repository = self.get_repository(org=org, repo=repo)
        return self._get_pull_requests(repository=repository)


class GitlabAPI:
    """Create and execute requests using the Gitlab API"""

    def __init__(self) -> None:
        self._client = Gitlab(url=config.GITLAB_URL, private_token=config.GITLAB_TOKEN)

    def get_repository(self, project_id: str, namespace: str) -> GitlabRepository:
        """Returns a repository for a given organization."""
        try:
            return self._client.projects.get(id=project_id)  # type: ignore
        except UnknownObjectException:
            raise RepositoryDoesNotExist(f"{project_id} does not exist for {namespace}")
        except BadCredentialsException:
            raise InvalidGitlabToken("GITLAB_TOKEN is not valid.")

    @staticmethod
    def _get_pull_requests(
        repository: Repository,
        state: str = "open",
        sort: str = "created",
    ) -> List[PullRequest]:
        return list(repository.get_pulls(state=state, sort=sort))

    def get_pull_requests(self, project_id: str, namespace: str) -> List[GitlabMergeRequest]:
        """Returns a list of pull requests for a given organization and repository."""
        repository = self.get_repository(project_id=project_id, namespace=namespace)  # type: ignore
        return repository.mergerequests.list(state="opened", order_by="created_at", sort="asc")  # type: ignore



from datetime import datetime
from time import sleep
from typing import List, Tuple

from rich.console import Group
from rich.live import Live
from rich.panel import Panel

from ..controller import GithubPullRequestController, GitlabPullRequestController, PullRequestController
from ..layout import (
    RenderLayoutManager,
    generate_layout,
    generate_log_table,
    generate_progress_tracker,
    generate_tree_layout,
)
from .controller import render_config_table
from .helpers import get_configuration
from .settings import (
    GITHUB_DEFAULT_PAGE_SIZE,
    GITHUB_TOKEN,
    GITHUB_URL,
    GITHUB_USER,
    GITLAB_TOKEN,
    GITLAB_URL,
    GITLAB_USER,
    REVIEWS_DELAY_REFRESH,
    REVIEWS_GITHUB_REPOSITORY_CONFIGURATION,
    REVIEWS_GITLAB_REPOSITORY_CONFIGURATION,
    REVIEWS_LABEL_CONFIGURATION,
    REVIEWS_PATH_TO_CONFIG,
)

logs: List[Tuple[str, str]] = []


def add_log_event(message: str) -> List[Tuple[str, str]]:
    """adds a log event to a list of logs and displays the top 20."""
    global logs

    logs = logs[-20:]
    logs.append((str(datetime.now().strftime("%Y-%m-%d %H:%M:%S")), f"[white]{message}"))
    return logs


def single_render(provider: str) -> None:
    """Renders the Terminal UI Dashboard once before closing the application"""

    configuration = None
    body = None

    if provider == "gitlab":
        configuration = get_configuration(config=REVIEWS_GITLAB_REPOSITORY_CONFIGURATION)
        body = GitlabPullRequestController().render(configuration=configuration)
    else:
        configuration = get_configuration(config=REVIEWS_GITHUB_REPOSITORY_CONFIGURATION)
        body = GithubPullRequestController().render(configuration=configuration)

    layout_manager = RenderLayoutManager(layout=generate_layout(log=False, footer=False))
    layout_manager.render_layout(
        progress_table=None,
        body=body,
        pull_request_component=generate_tree_layout(configuration=configuration),
        log_component=None,
    )

    with Live(
        renderable=layout_manager.layout,
        refresh_per_second=5,
        transient=False,
        screen=False,
    ):
        add_log_event(message="updated")


def render(provider: str) -> None:
    """Renders Terminal UI Dashboard"""
    (
        job_progress,
        overall_progress,
        overall_task,
        progress_table,
    ) = generate_progress_tracker()

    # initial load should be from database
    add_log_event(message="initializing...")

    configuration = None
    controller: PullRequestController

    if provider == "gitlab":
        configuration = get_configuration(config=REVIEWS_GITLAB_REPOSITORY_CONFIGURATION)
        controller = GitlabPullRequestController()
    else:
        configuration = get_configuration(config=REVIEWS_GITHUB_REPOSITORY_CONFIGURATION)
        controller = GithubPullRequestController()

    layout_manager = RenderLayoutManager(layout=generate_layout())
    layout_manager.render_layout(
        progress_table=progress_table,
        body=Panel(
            Group(),
            title="Activity",
            border_style="blue",
        ),
        pull_request_component=generate_tree_layout(configuration=configuration),
        log_component=generate_log_table(logs=logs),
    )

    with Live(layout_manager.layout, refresh_per_second=5, screen=True):
        while True:
            if not overall_progress or overall_progress.finished:
                (
                    job_progress,
                    overall_progress,
                    overall_task,
                    progress_table,
                ) = generate_progress_tracker()

            add_log_event(message="waiting...")

            # update view (blocking operation)
            layout_manager.render_layout(
                progress_table=progress_table,
                body=controller.render(configuration=configuration),
                pull_request_component=generate_tree_layout(configuration=configuration),
                log_component=generate_log_table(logs=logs),
            )

            delay = REVIEWS_DELAY_REFRESH * 0.01
            while not overall_progress.finished:
                sleep(delay)
                for job in job_progress.tasks:
                    if not job.finished:
                        job_progress.advance(job.id)

                completed = sum(task.completed for task in job_progress.tasks)
                overall_progress.update(overall_task, completed=completed)

            add_log_event(message="updated")


def render_config(show: bool) -> None:
    """Renders a table displaying configuration used by Reviews."""
    configurations = [
        {
            "name": "GITHUB_TOKEN",
            "value": GITHUB_TOKEN if show else "".join("*" for _ in range(len(GITHUB_TOKEN))),
        },
        {"name": "GITHUB_USER", "value": GITHUB_USER},
        {"name": "GITHUB_URL", "value": GITHUB_URL},
        {
            "name": "GITLAB_TOKEN",
            "value": GITLAB_TOKEN if show else "".join("*" for _ in range(len(GITLAB_TOKEN))),
        },
        {"name": "GITLAB_USER", "value": GITLAB_USER},
        {"name": "GITLAB_URL", "value": GITLAB_URL},
        {"name": "REVIEWS_PATH_TO_CONFIG", "value": f"{REVIEWS_PATH_TO_CONFIG}"},
        {
            "name": "GITHUB_DEFAULT_PAGE_SIZE",
            "value": f"{GITHUB_DEFAULT_PAGE_SIZE}",
        },
        {"name": "REVIEWS_DELAY_REFRESH", "value": f"{REVIEWS_DELAY_REFRESH}"},
        {
            "name": "REVIEWS_GITHUB_REPOSITORY_CONFIGURATION",
            "value": ", ".join(REVIEWS_GITHUB_REPOSITORY_CONFIGURATION),
        },
        {
            "name": "REVIEWS_GITLAB_REPOSITORY_CONFIGURATION",
            "value": ", ".join(REVIEWS_GITLAB_REPOSITORY_CONFIGURATION),
        },
        {
            "name": "REVIEWS_LABEL_CONFIGURATION",
            "value": ", ".join(REVIEWS_LABEL_CONFIGURATION),
        },
    ]

    render_config_table(configurations=configurations)



from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from typing import Dict, List, Tuple, Union

from github.PullRequest import PullRequest as ghPullRequest
from gitlab.v4.objects import ProjectMergeRequest as GitlabMergeRequest
from rich.console import Group
from rich.panel import Panel
from rich.table import Table

from reviews.source_control.client import GitlabAPI

from . import config
from .errors import RepositoryDoesNotExist
from .layout import render_pull_request_table, render_repository_does_not_exist
from .source_control import GithubAPI, Label, PullRequest


class PullRequestController:
    """Abstract base class for providier specific pull request controllers"""

    def render(self, configuration: List[Tuple[str, str]]) -> Panel:
        """Renders all pull requests for the provided configuration"""
        raise NotImplementedError("the render method needs to be implemented for your specified provider")


class GithubPullRequestController(PullRequestController):
    """retrieve and store pull requests."""

    def __init__(self) -> None:
        self.client = GithubAPI()

    def retrieve_pull_requests(self, org: str, repository: str) -> Union[Table, None]:
        """Renders Terminal UI Dashboard"""

        title = f"{org}/{repository}"
        pull_requests = []
        try:
            pull_requests = self.update_pull_requests(org=org, repository=repository)
        except RepositoryDoesNotExist:
            return render_repository_does_not_exist(
                title=title,
                link=f"https://www.github.com/{org}/{repository}",
            )

        if not pull_requests:
            return None

        return render_pull_request_table(
            title=title,
            pull_requests=pull_requests,
        )

    def update_pull_requests(self, org: str, repository: str) -> List[PullRequest]:
        """Updates repository models."""

        def _get_reviews(pull_request: ghPullRequest) -> Dict[str, str]:
            """Inner function to retrieve reviews for a pull request"""
            reviews = pull_request.get_reviews()
            res, seen = {}, []

            for review in reviews:
                if review.user.login in seen or review.state == "COMMENTED":
                    continue
                res[review.user.login] = review.state
                seen.append(review.user.login)
            return res

        pull_requests = self.client.get_pull_requests(org=org, repo=repository)

        code_review_requests = []
        for pull_request in pull_requests:

            reviews = _get_reviews(pull_request=pull_request)

            if pull_request.user.login == config.GITHUB_USER:
                approved_by_me = "AUTHOR"
            else:
                approved_by_me = reviews.get(config.GITHUB_USER, "")  # NOQA: R1721

            approved_by_others = any(
                [True for user, status in reviews.items() if user != config.GITHUB_USER and status == "APPROVED"]
            )
            labels = [Label(name=label.name) for label in pull_request.get_labels() if label.name]

            code_review_requests.append(
                PullRequest(
                    number=pull_request.number,
                    title=pull_request.title,
                    draft=pull_request.draft,
                    repository_url=f"https://www.github.com/{org}/{repository}",
                    link=f"https://www.github.com/{org}/{repository}/pull/{pull_request.number}",
                    additions=pull_request.additions,
                    deletions=pull_request.deletions,
                    created_at=pull_request.created_at.astimezone(tz=timezone.utc),
                    updated_at=pull_request.updated_at.astimezone(tz=timezone.utc),
                    author=pull_request.user.login,
                    approved=approved_by_me,
                    approved_by_others=approved_by_others,
                    labels=labels,
                )
            )

        return code_review_requests

    def render(self, configuration: List[Tuple[str, str]]) -> Panel:
        """Renders all pull requests for the provided configuration"""

        with ThreadPoolExecutor(max_workers=10) as executor:
            arguments = ((org, repo) for (org, repo) in configuration)
            tables = list(executor.map(lambda f: self.retrieve_pull_requests(*f), arguments))

        # filter unrenderable `None` results
        return Panel(
            Group(*[t for t in tables if t]),
            title="Activity",
            border_style="blue",
        )


class GitlabPullRequestController(PullRequestController):
    """retrieve and store pull requests."""

    def __init__(self) -> None:
        self.client = GitlabAPI()

    def retrieve_pull_requests(self, project_id: str, namespace: str) -> Union[Table, None]:
        """Renders Terminal UI Dashboard"""
        pull_requests = []
        try:
            pull_requests = self.update_pull_requests(project_id=project_id, namespace=namespace)
        except RepositoryDoesNotExist:
            return render_repository_does_not_exist(
                title=namespace,
                link=f"https://gitlab.com/{namespace}",
            )

        if not pull_requests:
            return None

        return render_pull_request_table(
            title=namespace,
            pull_requests=pull_requests,
        )

    def update_pull_requests(self, project_id: str, namespace: str) -> List[PullRequest]:
        """Updates repository models."""

        def _get_reviews(pull_request: GitlabMergeRequest) -> Dict[str, str]:
            """Inner function to retrieve reviews for a pull request"""
            reviews = pull_request.approvals.get()
            approvers = getattr(reviews, "approvers", [])

            return {approver["user"]["username"]: "approved" for approver in approvers} if approvers else {}

        # ProjectMergeRequest
        pull_requests = self.client.get_pull_requests(project_id=project_id, namespace=namespace)

        code_review_requests = []
        for pull_request in pull_requests:

            reviews = _get_reviews(pull_request=pull_request)

            if pull_request.author["username"] == config.GITLAB_USER:
                approved_by_me = "AUTHOR"
            else:
                approved_by_me = reviews.get(config.GITLAB_USER, "")  # NOQA: R1721

            approved_by_others = any(
                [True for user, status in reviews.items() if user != config.GITLAB_USER and status == "APPROVED"]
            )

            def get_labels(labels: List[str]) -> List[Label]:
                visible_labels = []
                labels_mapping = {label: len(label) for label in labels}

                current_size = 0
                max_size = 18
                hidden_labels = 0
                for label, size in labels_mapping.items():
                    if current_size == 0:
                        current_size += size
                        visible_labels.append(label)
                        continue

                    if current_size + size < max_size:
                        current_size += size
                        visible_labels.append(label)
                    else:
                        hidden_labels += 1

                labels = [Label(name=label) for label in visible_labels]

                if hidden_labels:
                    labels.append(Label(name=f"+{hidden_labels} others"))

                return labels

            labels = get_labels(labels=pull_request.labels)

            link = pull_request.web_url

            code_review_requests.append(
                PullRequest(
                    number=pull_request.iid,
                    title=pull_request.title,
                    author=pull_request.author["username"],
                    draft=pull_request.draft,
                    additions=0,
                    deletions=0,
                    link=link,
                    repository_url=link.split("/-/")[0],
                    created_at=datetime.strptime(pull_request.created_at, "%Y-%m-%dT%H:%M:%S.%f%z"),
                    updated_at=datetime.strptime(pull_request.updated_at, "%Y-%m-%dT%H:%M:%S.%f%z"),
                    approved=approved_by_me,
                    approved_by_others=approved_by_others,
                    labels=labels,
                )
            )

        return code_review_requests

    def render(self, configuration: List[Tuple[str, str]]) -> Panel:
        """Renders all pull requests for the provided configuration"""

        tables = [
            self.retrieve_pull_requests(project_id=project_id, namespace=namespace)
            for (project_id, namespace) in configuration
        ]

        # filter unrenderable `None` results
        return Panel(
            Group(*[t for t in tables if t]),
            title="Activity",
            border_style="blue",
        )



from ..config.helpers import get_configuration  # NOQA: F401
from ..config.settings import (  # NOQA: F401
    GITHUB_DEFAULT_PAGE_SIZE,
    GITHUB_TOKEN,
    GITHUB_URL,
    GITHUB_USER,
    GITLAB_DEFAULT_PAGE_SIZE,
    GITLAB_TOKEN,
    GITLAB_URL,
    GITLAB_USER,
    REVIEWS_AUTHOR,
    REVIEWS_DELAY_REFRESH,
    REVIEWS_GITHUB_REPOSITORY_CONFIGURATION,
    REVIEWS_GITLAB_REPOSITORY_CONFIGURATION,
    REVIEWS_LABEL_CONFIGURATION,
    REVIEWS_PATH_TO_CONFIG,
)



from datetime import datetime, timezone
from typing import Dict

import pytest
from freezegun import freeze_time

from reviews.source_control import PullRequest, Label


@pytest.fixture
def organization() -> str:
    return "my-org"


@pytest.fixture
def repository() -> str:
    return "my-repo"


@pytest.fixture
def label_colour_map() -> Dict[str, str]:
    return {"python": "[green]"}


@pytest.fixture
@freeze_time("2020-01-01T00:00:00+00:00")
def pull_request(organization, repository) -> PullRequest:
    return PullRequest(
        number=1,
        draft=True,
        author="me",
        repository_url=f"https://www.github.com/{organization}/{repository}",
        link=f"https://www.github.com/{organization}/{repository}/pull/1",
        title="[1] Initial Commit",
        created_at=datetime.now(timezone.utc),
        updated_at=datetime.now(timezone.utc),
        approved="",
        approved_by_others=False,
        labels=[Label(name="Python")],
    )


@pytest.mark.parametrize(
    "approved, expected",
    [
        ("", ""),
        ("APPROVED", "[green]Approved[/]"),
        ("CHANGES_REQUESTED", "[red]Changes Requested[/]"),
    ],
)
def test_pull_request_renders_approved(pull_request, approved, expected):
    pull_request.approved = approved
    rendered = pull_request.render_approved()
    assert rendered == expected


@pytest.mark.parametrize(
    "approved_by_others, expected",
    [
        (False, ""),
        (True, "[green]Ready[/]"),
    ],
)
def test_pull_request_renders_approved_by_others(pull_request, approved_by_others, expected):
    pull_request.approved_by_others = approved_by_others
    rendered = pull_request.render_approved_by_others()
    assert rendered == expected


@pytest.mark.parametrize(
    "labels, expected",
    [
        ([Label(name="Python")], "[green]Python[/]"),
        ([Label(name="Enhancement")], "[white]Enhancement[/]"),
        (
            [Label(name="Python"), Label(name="Enhancement")],
            "[green]Python[/], [white]Enhancement[/]",
        ),
    ],
)
def test_pull_request_renders_labels(pull_request, label_colour_map, labels, expected):
    pull_request.labels = labels
    rendered = pull_request.render_labels(label_colour_map)
    assert rendered == expected


@pytest.mark.parametrize(
    "title, draft, expected",
    [
        ("[1] Initial Commit", False, "[white]"),
        ("[Security][1] Initial Commit", False, "[bold red][Security][/][white]"),
        ("[1] Initial Commit", True, "[bold grey][Draft][/] [white]"),
        (
            "[Security][1] Initial Commit",
            True,
            "[bold red][Security][/][bold grey][Draft][/] [white]",
        ),
    ],
)
def test_pull_request_renders_title(pull_request, organization, repository, title, draft, expected):
    pull_request.draft = draft
    pull_request.title = title
    rendered = pull_request.render_title()
    assert rendered.startswith(expected)


@pytest.mark.parametrize(
    "since, expected",
    [
        (datetime(2020, 1, 2).astimezone(tz=timezone.utc), "[yellow]a day ago[/]"),
        (datetime(2020, 1, 8).astimezone(tz=timezone.utc), "[red]7 days ago[/]"),
    ],
)
def test_pull_request_renders_updated_at(pull_request, since, expected):
    rendered = pull_request.render_updated_at(since=since)
    assert rendered == expected



from operator import attrgetter
from typing import Dict, List, Tuple

from rich import box
from rich.color import ANSI_COLOR_NAMES
from rich.console import Group
from rich.layout import Layout
from rich.panel import Panel
from rich.progress import BarColumn, Progress, SpinnerColumn, TaskID, TextColumn
from rich.table import Table
from rich.tree import Tree

from ..config import settings
from ..source_control import PullRequest


def get_label_colour_map() -> Dict[str, str]:
    """converts a comma separated list of organizations/repositories into a list
    of tuples.
    """

    def _preproc(label_colour: str) -> List[str]:
        return label_colour.lower().split(sep="/")

    return {
        label: f"[{colour}]"
        for label, colour in map(_preproc, settings.REVIEWS_LABEL_CONFIGURATION)
        if colour in ANSI_COLOR_NAMES
    }


def render_repository_does_not_exist(
    title: str,
    link: str,
) -> Table:
    """Renders a list of pull requests as a table"""
    table = Table(show_header=True, header_style="bold white")
    table.add_column("#", style="dim", width=7)
    table.add_column(
        f"[link={link}]{title}[/link]",
        width=160,
    )

    table.add_row(
        "",
        "Please confirm this repository exists and that you can access it before attempting to use it.",
    )

    return table


def render_pull_request_table(
    title: str,
    pull_requests: List[PullRequest],
) -> Table:
    """Renders a list of pull requests as a table"""

    show_diff = False
    show_author = settings.REVIEWS_AUTHOR

    if pull_requests and pull_requests[0].repository_url:
        link = f"[link={pull_requests[0].repository_url}]{title}[/link]"
    else:
        link = f"{title}"

    table = Table(show_header=True, header_style="bold white")
    table.add_column("#", style="dim", width=5)
    table.add_column(link, width=55)

    if show_author:
        table.add_column("Author", width=20)

    table.add_column("Labels", width=30)
    table.add_column("Diff +/-", width=10)
    table.add_column("Activity", width=15)
    table.add_column("Approved", width=10)
    table.add_column("Mergeable", width=10)

    label_colour_map = get_label_colour_map()

    for pr in sorted(pull_requests, key=attrgetter("updated_at"), reverse=True):
        row = [
            f"[white]{pr.number} ",
            pr.render_title(),
        ]

        if show_author:
            row.append(pr.render_author())

        row.append(pr.render_labels(label_colour_map))

        if show_diff:
            row.append(pr.render_diff())
        else:
            row.append("    -    ")

        row.append(pr.render_updated_at())
        row.append(pr.render_approved())
        row.append(pr.render_approved_by_others())

        table.add_row(*row)

    return table


def generate_layout(log: bool = True, footer: bool = True) -> Layout:
    """Define the layout for the terminal UI."""
    layout = Layout(name="root")

    sections = [Layout(name="header", size=3), Layout(name="main", ratio=1)]
    if footer:
        sections.append(Layout(name="footer", size=7))
    layout.split(*sections)

    layout["main"].split_row(  # type: ignore
        Layout(name="left_side", size=25),
        Layout(name="body", ratio=2, minimum_size=90),
    )

    nav_sections = [Layout(name="configuration")]
    if log:
        nav_sections.append(Layout(name="log"))

    layout["left_side"].split(*nav_sections)

    return layout


def generate_tree_layout(configuration: List[Tuple[str, str]]) -> Group:
    """Generates a tree layout for the settings configuration"""
    organization_tree_mapping: Dict[str, Tree] = {}
    for (org, repo) in configuration:
        tree = organization_tree_mapping.get(f"{org}", Tree(f"[white]{org}"))
        tree.add(f"[link=https://www.github.com/{org}/{repo}]{repo}[/link]")
        organization_tree_mapping[org] = tree

    return Group(*organization_tree_mapping.values())


def generate_log_table(logs: List[Tuple[str, str]]) -> Table:
    """Generetes a table for logging activity"""
    table = Table("Time", "Message", box=box.SIMPLE)

    if logs:
        for log in logs:
            time, message = log
            table.add_row(time, message)

    return table


def generate_progress_tracker() -> Tuple[Progress, Progress, TaskID, Table]:
    """Tracks the progress of tasks"""
    progress = Progress(
        "{task.description}",
        SpinnerColumn(),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    )
    progress.add_task("[white]Pull Requests", total=100)

    total = sum(task.total for task in progress.tasks if isinstance(task.total, (int, float)))
    overall_progress = Progress()
    overall_task = overall_progress.add_task(description="All", total=int(total))

    progress_table = Table.grid(expand=True)
    progress_table.add_row(
        Panel(
            renderable=overall_progress,  # type: ignore
            title="Next Refresh",
            border_style="blue",
        ),
        Panel(
            renderable=progress,  # type: ignore
            title="[b]Next fetch for:",
            border_style="blue",
            padding=(1, 2),
        ),
    )

    return progress, overall_progress, overall_task, progress_table



from typing import Dict, List, Optional

import humanize
from github.Repository import Repository
from rich.console import Console
from rich.table import Table

from ..config import REVIEWS_GITHUB_REPOSITORY_CONFIGURATION, get_configuration
from ..errors import RepositoryDoesNotExist
from ..source_control import GithubAPI


def repository_metrics() -> None:
    """Renders an aggregated table view displaying contributors by contributors
    for the lifetime of a repository.
    """
    console = Console()

    for org, repo in get_configuration(config=REVIEWS_GITHUB_REPOSITORY_CONFIGURATION):
        client = GithubAPI()
        repository: Optional[Repository] = None
        try:
            repository = client.get_repository(org=org, repo=repo)
        except RepositoryDoesNotExist:
            continue

        console.print()
        table = Table(title=f"Repository Activity for {org}/{repo}")
        table.add_column("User", no_wrap=True, width=25)
        table.add_column("Additions", width=10)
        table.add_column("Deletions", width=10)
        table.add_column("Commits", width=10)

        if repository:
            for contributor in repository.get_stats_contributors():  # type: ignore
                name = contributor.author.name
                if not name:
                    continue

                additions = 0
                deletions = 0
                commits = 0
                for week in contributor.weeks:
                    additions += week.a
                    deletions += week.d
                    commits += week.c

                additions = f"[green]+{humanize.intcomma(additions)}[/]"
                deletions = f"[red]-{humanize.intcomma(deletions)}[/]"
                commits = f"[white]{humanize.intcomma(commits)}[/]"

                table.add_row(name, additions, deletions, commits)

        console.print(table)
        console.print()


def render_config_table(configurations: List[Dict[str, str]]) -> None:
    """Renders a table using the supplied configuration."""
    table = Table()
    table.add_column("Name", style="white", no_wrap=True)
    table.add_column("Value", style="cyan")

    for configuration in configurations:
        table.add_row(configuration["name"], configuration["value"])

    console = Console()
    console.print(table)



class RepositoryDoesNotExist(Exception):
    """Unable to fetch the repository as it already exists"""


class InvalidGithubToken(Exception):
    """Unable to query Github using an empty or invalid Github Token"""


class InvalidGitlabToken(Exception):
    """Unable to query Gitlab using an empty or invalid Gitlab Token"""
